/************************************************************************
    frap/defrap - fractal image compressor/decompressor
    Copyright (C) 1997-2000 Matthias Ruhl

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    The author can be contacted at:
      MIT Laboratory for Computer Science
      200 Technology Square
      Cambridge, MA 02139
      USA
    or via e-mail: ruhl@theory.lcs.mit.edu
*************************************************************************/

#include <stdlib.h>
#include <math.h>
#include "frap.h"

static int compare_codeT(const void *c1, const void *c2)
{
  double err1 = ((const codeT *) c1)->err;
  double err2 = ((const codeT *) c2)->err;
  
  if(err1 < err2) return -1;
  return (err1 > err2);
}

void rd_compare(rangeT *r, int keep)
{
  int i,j;
  
  double rsum,rsum2,dsum,dsum2,rdsum;
  double scaling,offset,det,n;
  int qscaling,qoffset,sym_op;
  int rbeg,rend;
  int neg,dx,dy;
  int cnum;
  codeT *t;

  n = basesize * basesize * r->size;

  for(cnum=0;cnum<r->codenum;cnum++)
    {
      t = r->code+cnum;
      
      sym_op = t->rot;
      neg = sym_op&4;
      dx = 1 - (sym_op&2);
      dy = 1 -((sym_op&1)<<1);

      if(t->precomp)
	{
	  dsum  = t->dsum;
	  dsum2 = t->dsum2;
	  rsum  = t->rsum;
	  rsum2 = t->rsum2;
	  rdsum = t->rdsum;
	}
      else
	dsum = dsum2 = rsum = rsum2 = rdsum = 0.0;
      if(t->precomp >= 0)
	{
	  rbeg = t->precomp;
	  rend = r->size;
	}
      else
	{
	  rbeg = 0;
	  rend = -(t->precomp);
	}
      for(i=rbeg;i<rend;i++)
	{
	  int x,y,id;

	  if(neg == 0)
	    {
	      x = (width2*height2  + t->domain%(width2) + dx*(r->pos[i]%width-r->pos[0]%width))%(width2);
	      y = (width2*height2 + t->domain/(width2) + dy*(r->pos[i]/width-r->pos[0]/width))%(height2);
	    }
	  else
	    {
	      x = (width2*height2 + t->domain%(width2) + dy*(r->pos[i]/width-r->pos[0]/width))%(width2);
	      y = (width2*height2 + t->domain/(width2) + dx*(r->pos[i]%width-r->pos[0]%width))%(height2);
	    }

	  id = y*(width2)+x;

	  dsum += csum[id];
	  dsum2 += cvar[id];
	  rsum += bsum[r->pos[i]];
	  rsum2 += bvar[r->pos[i]];

	  for(j=0;j<basesize*basesize;j++)
	    rdsum += codebook[id][rot[sym_op][j]]*baseblock[r->pos[i]][j];
	}

      t->dsum = dsum;
      t->dsum2 = dsum2;
      t->rsum = rsum;
      t->rsum2 = rsum2;
      t->rdsum = rdsum;
      t->precomp = r->size;

      det = n*dsum2 - dsum*dsum;
      
      if (det == 0.0)
	scaling = 0.0;
      else
	scaling = (n*rdsum - dsum*rsum)/det;
      
      qscaling = (int) rint((scaling + 1.0)/2.0*(double)(1<<s_bits));
      if (qscaling < 0) qscaling = 0;
      if (qscaling >= 1<<s_bits) qscaling = (1<<s_bits)-1;
      
      scaling = (double)qscaling/(double)(1<<s_bits)*2.0-1.0;
      offset = (rsum - scaling*dsum)/n;
      
      if (scaling > 0.0)  offset += scaling*255;
      qoffset = (int) rint(offset/((1.0+fabs(scaling))*255)*(double)((1<<o_bits)-1));
      if (qoffset < 0) qoffset = 0;
      if (qoffset >= 1<<o_bits) qoffset = (1<<o_bits)-1;
      
      offset = (double)qoffset/(double)((1<<o_bits)-1)*((1.0+fabs(scaling))*255);
      if (scaling > 0.0) offset -= scaling*255;
      
      t->err = rsum2 + scaling*(scaling*dsum2 - 2.0*rdsum + 2.0*offset*dsum) +
	offset*(offset*n - 2.0*rsum);
      t->qscaling = qscaling;
      t->qoffset  = qoffset;
    }
  qsort(r->code,r->codenum,sizeof(codeT),compare_codeT);
  if(keep < 1) keep = r->codenum; /* keep everything */
  if(keep < r->codenum)
    {
      codeT *tmp = (codeT *) malloc(keep * sizeof(codeT));

      for(i=0;i<keep;i++) tmp[i] = r->code[i];
      free(r->code);
      r->code = tmp;
      r->codenum = keep;
    }
  r->timestamp = gettime();
}
